"""

"""
import matplotlib
import matplotlib.pyplot as plt
# from ctypes import windll

# sdpi = windll.user32.GetDpiForSystem()

plt.style.use('bmh')
matplotlib.rcParams['axes.prop_cycle'] = matplotlib.cycler('color', ['E24A33', '348ABD', '988ED5', '8EBA42', 'FBC15E', '777777', 'FFB5B8'])
                   # E24A33 : red
                   # 348ABD : blue
                   # 988ED5 : purple
                   # 8EBA42 : green
                   # FBC15E : yellow
                   # 777777 : gray
                   # FFB5B8 : pink

# matplotlib.rcParams['figure.dpi'] = sdpi
matplotlib.rcParams['savefig.dpi'] = 800
matplotlib.rcParams['font.size'] = 10
matplotlib.rcParams['markers.fillstyle'] = 'none'
matplotlib.rcParams["errorbar.capsize"] = 3

colors = ['#E24A33', '#348ABD', '#988ED5', '#8EBA42', '#FBC15E', '#777777', '#FFB5B8']